/*
 * BranchDefaultMutableTreeNode.java
 *
 * Created on January 22, 2006, 3:45 PM
 *
 *
Copyright (c) 2006 Forward Computing and Control Pty. Ltd.
NSW. Australia,  www.forward.com.au
All rights reserved.
 
Redistribution of the source of this package, with or without
modification, is NOT permitted.
 
Redistribution of the compiled java .class files of this package, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions in compiled java .class form must reproduce the above
   copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the distribution.
 
2. If the software is modified, the compiled java .class distribution must
   include, in the documentation and/or other materials provided with the
   distribution, a statement that the software has been modified and
   identify the person who was responsable for the modifications.
 
3. Neither the name of the author nor Forward Computing and Control may be
   used to endorse or promote products derived from this software
   without specific prior written permission.
 
4.  All advertising materials mentioning features or use of this
    software must display the following acknowledgment:
    "This product includes software developed by Matthew Ford
     and Forward Computing and Control Pty. Ltd.
    (http://www.forward.com.au)."
 
 
THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


package au.com.forward.undoRedo;

import java.util.Enumeration;
import javax.swing.undo.UndoableEdit;

/**
 *
 * @author Matthew Ford
 */
/**
 * A mutable tree node to hold Redo branches.
 * <br> Overrides toString() to indicate if any node in this branch have been redone.
 * <br> Added updateDisplayMessage() to update the node display string to show the
 * deepest redo in this branch.
 */
public class BranchDefaultMutableTreeNode extends UndoRedoTreeNode implements IHaveDescription {
    
    private int maxDepth = 0;
    private UndoRedoTreeNode deepestNode = null;
    
    /**
     * BranchDefaultMutableTreeNode constructor
     * @param userObject the BranchUndoRedo
     */
    public BranchDefaultMutableTreeNode(BranchUndoRedo userObject) {
        super(userObject);
    }
    
    /**
     * Checks the first child of this branch to see if it has been redone.
     * <br>All UndoRedo objects in a branch are Redo objects, that is isUndo()==FALSE
     * but this checks the underlying UndoableEdit to see if it has been redone
     * or is dead
     *
     * @param node the branch node to check
     * @return 0 if first child of this branch is dead<br>
     *  1 if first child of this branch is a Redo that has been redone<br>
     *  -1 if first child of this branch is a Redo that has not been redone
     */
    protected static int isFirstChildReDone(BranchDefaultMutableTreeNode node) {
        if (node.getChildCount()>0) {
            UndoRedoTreeNode firstChild = (UndoRedoTreeNode)node.getFirstChild();
            if (firstChild instanceof BranchDefaultMutableTreeNode) {
                return isFirstChildReDone((BranchDefaultMutableTreeNode)firstChild);
            } // else
            UndoRedo undoRedo = (UndoRedo)firstChild.getUserObject();
            if (undoRedo.isDead()) {
                return 0;
            } else if ((!undoRedo.canRedo()) && (!undoRedo.isUndo())) {
                return 1;
            } else {
                return -1;
            }
        } // else
        // should always have a child for a branch node
        return 0;
    }
    
    /**
     * Override toString() to indicate if any nodes in this branch have been redone
     * @return super.toString with prefix if child has been redone.
     */
    public String toString() {
        String rtn = super.toString();
        // see if deepest child has been re-done
        if ( (getDeepestNode() != null) &&
                !((UndoRedo)getDeepestNode().getUserObject()).canRedo()) {
            rtn = UndoRedo.HTML_START+UndoRedo.escapeHTML(UndoRedo.ALL_REDONE_PREFIX) + UndoRedo.escapeHTML(rtn);
        }  else  if (isFirstChildReDone(this) > 0) {
            // first child has been re-done
            rtn = UndoRedo.HTML_START+UndoRedo.escapeHTML(UndoRedo.REDONE_PREFIX) + UndoRedo.escapeHTML(rtn);
        }
        return rtn;
    }
    
    public UndoRedoTreeNode getDeepestNode() {
        return deepestNode;
    }
    
    public void setDeepestNode(UndoRedoTreeNode deepestNode) {
        this.deepestNode = deepestNode;
    }
    
    /**
     * Get the maximum depth of this branch.
     * @return the maximum depth.
     */
    public int getMaxDepth() {
        return maxDepth;
    }
    
    /**
     * Set the maximum depth of this branch
     * @param maxDepth the maximum depth
     */
    public void setMaxDepth(int maxDepth) {
        this.maxDepth = maxDepth;
    }
    
    /**
     * Gets the Description of the UndoableEdit.
     * Forwards request to the deepest node
     * @param textDisplay - the callback component used to display the extended description
     */
    public void getDescription(IDisplayText textDisplay, UndoRedoTreeNode undoRedoTreeNode) {
        if (undoRedoTreeNode != this) {
            throw new IllegalArgumentException(" UndoRedoTreeNode argument is not the node for this BranchNode.");
        }         
        undoRedoTreeNode = getDeepestNode();
        UndoRedo undoRedo = (UndoRedo)undoRedoTreeNode.getUserObject();
        UndoableEdit edit = undoRedo.getUndoableEdit();
        if (edit instanceof IHaveDescription) {
            ((IHaveDescription)edit).getDescription(textDisplay, undoRedoTreeNode);
        }
    }
    
    /**
     * Cancels getDescription
     * Forwards request to the deepest node
     */
    public void cancelGetDescription() {
        UndoRedoTreeNode undoRedoTreeNode = getDeepestNode();
        UndoableEdit edit = ((UndoRedo)undoRedoTreeNode.getUserObject()).getUndoableEdit();
        if (edit instanceof IHaveDescription) {
            ((IHaveDescription)edit).cancelGetDescription();
        }
    }
    
    /**
     * Find the deepest presentation name and set it as this nodes presentation name
     */
    public void updateDisplayMessage() {
        // get the maximum number of children and the deepest presentation name
        int nodeCount = 0;
        int depth = 0;
        UndoRedoTreeNode deepestNode = null;
        String presentationName = "";
        Enumeration<UndoRedoTreeNode> ce = children();
        while(ce.hasMoreElements()) {
            UndoRedoTreeNode cn = ce.nextElement();
            if (cn instanceof BranchDefaultMutableTreeNode) {
                BranchDefaultMutableTreeNode branchTreeNode = (BranchDefaultMutableTreeNode)cn;
                BranchUndoRedo subBranch = (BranchUndoRedo)(cn.getUserObject());
                if ((branchTreeNode.getMaxDepth()+nodeCount) > depth) {
                    depth = branchTreeNode.getMaxDepth();
                    presentationName = subBranch.getBranchSuffix();
                    deepestNode = branchTreeNode.getDeepestNode();
                }
            } else {
                // normal node
                UndoRedo subNode = (UndoRedo)cn.getUserObject();
                nodeCount++;
                if (nodeCount >= depth) {
                    depth = nodeCount;
                    presentationName = subNode.getRedoPresentationName();
                    deepestNode = cn;
                }
            }
        }
        BranchUndoRedo branchUndoRedo = (BranchUndoRedo)getUserObject();
        branchUndoRedo.setBranchSuffix(presentationName);
        setMaxDepth(depth);
        setDeepestNode(deepestNode);
    }
}
