/*
 * ICanBeRolledBack.java
 *
 * Created on April 13, 2005, 8:51 PM
 *
 Copyright (c) 2005 Forward Computing and Control Pty. Ltd.
NSW. Australia,  www.forward.com.au
All rights reserved.
 
Redistribution of the source of this package, with or without
modification, is NOT permitted.
 
Redistribution of the compiled java .class files of this package, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions in compiled java .class form must reproduce the above
   copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the distribution.
 
2. If the software is modified, the compiled java .class distribution must
   include, in the documentation and/or other materials provided with the
   distribution, a statement that the software has been modified and
   identify the person who was responsable for the modifications.
 
3. Neither the name of the author nor Forward Computing and Control may be
   used to endorse or promote products derived from this software
   without specific prior written permission.
 
4.  All advertising materials mentioning features or use of this
    software must display the following acknowledgment:
    "This product includes software developed by Matthew Ford
     and Forward Computing and Control Pty. Ltd.
    (http://www.forward.com.au)."
 
 
THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package au.com.forward.undoRedo;

/**
 * This interface marks an UndoableEdit as being able
 * to roll back the undo/redo tree in the event that it fails or is cancelled.
 * This interface also allows for re-syncing the undo/redo tree
 * to the last edit sucessfully completed.
 *<p>
 * When the UndoableEdit is registered with the UndoRedoManager
 * (via {@link UndoRedoManager#undoableEditHappened}) and each
 * time and undo or redo is performed by the UndoRedoManager,
 * the {@link #setFrameMarker} is called with the
 * current frameMarker.
 *<p>
 * This means that the same UndoableEdit will have multiple
 * frameMarkers set as undos and redos are preformed. This means
 * you need to save the frameMarker somewhere else (typically 
 * with the task preforming the operation) at the beginning of
 * the intial edit and at the start of any undo/redo.
 * If the edit, undo or redo fails to complete sucessfully 
 * (i.e is cancelled or fails), call {@link UndoRedoManager#cancel}
 * to remove the unsucessful operation from the undo/redo tree.
 * You can also re-sync the tree to the last sucessfull command 
 * by calling {@link UndoRedoManager#syncToLastSuccessfull} passing it
 * the frame of the last command sucessfully processed.
 * If the edit, undo or redo is sucessful you can call
 * {@link UndoRedoManager#disposeOfRollBack} to free the resources
 * used by the frameMarker.  Alternatively freeing references
 * to the marker will allow it to be garbaged collected as 
 * it is only weekly held be the UndoRedoManager.
 *<p>
 *
 *
 * @author Matthew Ford
 */
public interface ICanBeRolledBack {
  
  /**
   * Called by the UndoRedoManager to set
   * the frame marker for future roll back.
   * <br>This method is called when the UndoableEdit is
   * registered with the UndoRedoManager and each time an
   * undo or redo operation is performed on this UndoableEdit,
   * so the frame marker it needs to be
   * saved elsewhere (usually with the task performing the operation).
   *
   * @param marker the frameMarker associated with this operation 
   */
  public void setFrameMarker(Long marker);
  
  /**
   * Returns the last frameMarker set.
   * <br>Because the frameMarker is set for the
   * original edit and every subsequent undo/redo
   * of it, the frameMarker needs to be
   * saved with the task preforming the operation.
   * <p>
   * A typical implementation of this method nulls the frameMarker when
   * it is returned.
   * <br> Nulling the frameMarker here allows it to be garbaged collected
   * from the UndoRedoManager when there are no other
   * references.
   *<br>I.e. a typical implementation would be:-
   * <pre>
   *  public Long getFrameMarker() {
   *   Long rtn = frameMarker;
   *   frameMarker = null;
   *   return rtn;
   *  }
   * </pre>
   * @return the last frameMarker set.
   */
  public Long getFrameMarker();

  /**
   * Sets the undo or redo state.
   * <br>Called multiple times by the UndoRedoManager when rolling back to
   * reset the UndoableEdit to its previous state.
   * <p>
   * This method should also do any other clean up necessary
   * in the UndoableEdit to return it to its previous state.
   * <p>This method is only called to toggle the state. That is
   * if the current state is canRedo() then
   * <br>setUndoRedoState(true);
   *<br> will be called to change the state to canUndo();
   *
   * <p>
   * Note: The <code>hasBeenDone</code> variable in AbstractUndoableEdit
   * has package scope and no direct setter, so if you are
   * extending from AbstractUndoableEdit 
   * then to set the hasBeenDone flag to true you need to call 
   * <pre>
   * try {
   *   super.undo();
   * } catch (CannotUndoException cex) {
   *  // ignore
   * }
   * </pre>
   * To set it to false use<pre>
   * try {
   *   super.redo();
   * } catch (CannotRedoException cex) {
   *  // ignore
   * }
   * </pre>
   *
   * @param undoState true to return UndoableEdit to undo state,
   * false to return it to a redo state.
   */
  public void setUndoState(boolean undoState);
  
  
}
