/*
 * UndoRedo.java
 * Created on March 8, 2005, 7:43 AM by Matthew Ford
 *
Copyright (c) 2005 Forward Computing and Control Pty. Ltd.
NSW. Australia,  www.forward.com.au
All rights reserved.
 
Redistribution of the source of this package, with or without
modification, is NOT permitted.
 
Redistribution of the compiled java .class files of this package, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions in compiled java .class form must reproduce the above
   copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the distribution.
 
2. If the software is modified, the compiled java .class distribution must
   include, in the documentation and/or other materials provided with the
   distribution, a statement that the software has been modified and
   identify the person who was responsable for the modifications.
 
3. Neither the name of the author nor Forward Computing and Control may be
   used to endorse or promote products derived from this software
   without specific prior written permission.
 
4.  All advertising materials mentioning features or use of this
    software must display the following acknowledgment:
    "This product includes software developed by Matthew Ford
     and Forward Computing and Control Pty. Ltd.
    (http://www.forward.com.au)."
 
 
THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
package au.com.forward.undoRedo;

import java.util.Enumeration;
import java.util.Vector;
import javax.swing.tree.TreeNode;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoableEdit;

/**
 * Wrapps an UndoableEdit for adding to the UndoRedoManager tree.
 * <br>Wrapping allows us to reuse the original UndoableEdit when
 * adding Undo nodes generated from redo branches. This avoids
 * the need to clone the original UndoableEdit, which may not be cloneable
 * or may require significant memory.
 *
 * @author Matthew Ford
 */
public class UndoRedo  implements UndoableEdit, Cloneable {
  
  /**
   * the wrapped edit
   */
  protected UndoableEdit edit;
  
  /**
   * Is this an Undo or a Redo
   */
  private boolean undo = true;  // no setter
  
  /**
   * Is this undo unreachable.
   *<br>Ususally due to an intervening Dead action
   */
  private boolean unreachable = false;

  /**
   * Prefix for HTML
   */
  public static String HTML_START = "<HTML>";
  
  /**
   * Prefix for redone nodes
   */
  public static String REDONE_PREFIX = " > ";
  
  /**
   * Prefix for all nodes redone. Used for branch nodes
   */
  public static String ALL_REDONE_PREFIX = " >> ";
  
  /**
   * Prefix for dead nodes
   */
  public static String DEAD_FORMAT = "<STRIKE>";
  
  /**
   * Prefix for unreachable nodes
   */
  public static String UNREACHABLE_FORMAT = "<STRIKE><FONT COLOR=\"GRAY\">";
  
  /**
   * Prefix for insignificant nodes
   */
  public static String INSIGNIFICANT_FORMAT = "<FONT COLOR=\"GRAY\">";
  
  /**
   * Creates a new instance of UndoRedo wrapping an UndoableEdit.
   * This is always an undo, (ie. undo == true)
   */
  public UndoRedo(UndoableEdit edit) {
    if (edit == null) {
      throw new IllegalArgumentException("UndoableEdit cannot be null.");
    }
    this.edit = edit;
  }
  
  
  /**
   * Returns the UndoableEdit wrapped by this UndoRedo
   * @return the UndoableEdit that is wrapped by this UndoRedo
   */
  public UndoableEdit getUndoableEdit() {
    return edit;
  }
  
  /**
   * Returns true if the obj is the complement of this UndoRedo.
   * UndoRedo complements wrap the same UndoableEdit object but
   * have complementary undo settings
   * @param obj  the object to test
   * @return true if the object is the complement of this UndoRedo
   */
  public boolean equalsComplement(Object obj) {
    if ((obj == null) || (!(obj instanceof UndoRedo) )) {
      return false;
    }
    UndoRedo complement = (UndoRedo)obj;
    if (edit != complement.edit) {
      return false;
    }
    if (undo == complement.undo){
      return false;
    } // undo == ! complement.undo
    return true;
  }
  
  
  /**
   * Returns the clone of this UndoRedo.
   * <br>Does not clone the wrapped UndoableEdit.
   * @throws java.lang.CloneNotSupportedException
   * @return a clone of this UndoRedo, with a reference to the original wrapped edit.
   */
  public Object clone() throws CloneNotSupportedException {
    UndoRedo undoRedo = (UndoRedo)super.clone();
    // leave edit as reference
    return undoRedo;
  }
  
  /**
   * Returns true if this is the undo version.
   * <br>This setting is independent of the undo/redo setting of the wrapped UndoableEdit.
   * @return true if Undo, false if Redo
   */
  public boolean isUndo() {
    return undo;
  }
  
  /**
   * Returns true if thie undo/redo cannot be reached.
   *<br> Usually due to an intervening dead node.
   * @return true if this undo/redo cannot be reached.
   */
  public boolean isUnreachable() {
    return unreachable;
  }
  
  /**
   * Sets this undo/redo as being un-reached.
   *<br> Usually due to an intervening dead node.
   *<br> Like die(), this setting cannot be undone.
   */
  public void setUnreachable() {
    this.unreachable = true;
  }
  
  /**
   * Returns the UndoRedo complement of this object.
   * <br>Keeps the original reference to wrapped Undoable edit.
   * @return  the UndoRedo complement of this object.
   */
  public UndoRedo getComplement() {
    UndoRedo undoRedo = null;
    try {
      undoRedo = (UndoRedo)clone();
    } catch (CloneNotSupportedException ce) {
      throw new RuntimeException("Clone not supported!!",ce);
    }
    undoRedo.undo = !undoRedo.undo;
    return undoRedo;
  }
  
  
  /**
   * Undo the edit that was made.
   * @throws javax.swing.undo.CannotUndoException  if cannot be undone.
   */
  public void undo() throws CannotUndoException {
    edit.undo();
  }
  
  /**
   * True if it is still possible to undo this operation.
   * @return true if this is an Undo UndoRedo and the wrapped UndoableEdit can be undone.
   */
  public boolean canUndo() {
    return undo && edit.canUndo();
  }
  
  /**
   * Re-apply the edit, assuming that it has been undone.
   * @throws javax.swing.undo.CannotRedoException  if cannot be redone.
   */
  public void redo() throws CannotRedoException {
    edit.redo();
  }
  
  /**
   * True if it is still possible to redo this operation.
   * @return true if this is a Redo UndoRedo and the wrapped UndoableEdit can be redone.
   *  false if either this is and Undo leaf or a Redo leaf wrapping an Undo edit
   */
  public boolean canRedo() {
    return (!undo) && edit.canRedo();
  }
  
  
  /**
   * Is the wrapped UndoableEdit dead
   * @return true if the wrapped UndoableEdit is dead, else false.
   */
  public boolean isDead() {
    return ((!edit.canRedo()) && (!edit.canUndo()));
  }
  
  /**
   * May be sent to inform an edit that it should no longer be
   * used. This is a useful hook for cleaning up state no longer
   * needed once undoing or redoing is impossible.
   * <br>The UndoRedoManager never calls this and the will throw an exception if it encounters a dead UndoRedo.
   * <br>Note that this is a one-way operation. There is no "un-die"
   * method.
   *
   */
  public void die() {
    edit.die();
  }
  
  /**
   * Does nothing, always returns false.
   * @param anEdit the edit to be added
   * @return always false
   */
  public boolean addEdit(UndoableEdit anEdit) {
    return false;
  }
  
  /**
   * Does nothing, always returns false.
   * @param anEdit the edit to be replace this edit
   * @return always false
   */
  public boolean replaceEdit(UndoableEdit anEdit) {
    return false;
  }
  
  /**
   * Returns false if this edit is insignificant--for example one
   * that maintains the user's selection, but does not change any
   * model state. This status can be used by an
   * <code>UndoableEditListener</code>
   * (like UndoManager) when deciding which UndoableEdits to present
   * to the user as Undo/Redo options, and which to perform as side
   * effects of undoing or redoing other events.
   * <p>UndoRedoManager does not take any notice of this hint.
   * @return always false
   */
  public boolean isSignificant() {
    return edit.isSignificant();
  }
  
  /**
   * Provides a localized, human readable description of this edit
   * , should be detailed enough to provide a the basis of a meaningful node lable.
   * <br>If name starts with <HTML> assume rest of string has been escaped already
   * @return the presentation name.
   */
  public String getPresentationName() {
    return edit.getPresentationName();
  }
  
  /**
   * Provides a localized, human readable description of the undoable
   * form of this edit.
   * <br>If name starts with <HTML> assume rest of string has been escaped already
   */
  public String getUndoPresentationName() {
    return edit.getUndoPresentationName();
  }
  
  /**
   * Provides a localized, human readable description of the redoable
   * form of this edit.
   * <br>If name starts with <HTML> assume rest of string has been escaped already
   */
  public String getRedoPresentationName() {
    return edit.getRedoPresentationName();
  }
  
  
  /**
   * Returns a string representation of this UndoRedo object.
   * <br>Use by the tree in its node display.
   *
   * @return a string representation of this object.
   */
  public String toString() {
    String prefix = "";
    String format = "";
    if (isDead() || isUnreachable()) {
      if (isDead()) {
        format = DEAD_FORMAT;
      }
      if (isUnreachable()) {
        format = UNREACHABLE_FORMAT;
      }
    } else if (!isSignificant()) {
      format = INSIGNIFICANT_FORMAT;
    }
    String presentationName = "";
    if (undo) {
      presentationName = getUndoPresentationName();
    } else {
      // this is a redo
      if (!isDead()) {
        prefix = (!canRedo()?REDONE_PREFIX:"");
      }
      presentationName = getRedoPresentationName();
    }
    return HTML_START+format+escapeHTML(prefix)+escapeHTML(presentationName);
  }
  
  public static String escapeHTML(String input) {
    if (input == null) {
      input = "";
    }
    if (input.toUpperCase().startsWith(HTML_START)) {
      // just remove <HTML> and assume the rest is escaped
      return input.substring(HTML_START.length());
    }
    StringBuffer buf = new StringBuffer(input);
    // handles nulls
    for (int i = 0; i < buf.length(); i++) {
      char c = buf.charAt(i);
      // Escape '&' and '<' and '>' and '"' and ' for well-formed Xml output.
      if (c == '<') {
        buf.setCharAt(i, '&');
        buf.insert(i + 1, "lt;");
        i += 3;
        // the extra charaters
      } else if (c == '&') {
        buf.setCharAt(i, '&');
        buf.insert(i + 1, "amp;");
        i += 4;
        // the extra charaters
      } else if (c == '>') {
        buf.setCharAt(i, '&');
        buf.insert(i + 1, "gt;");
        i += 3;
        // the extra charaters
      } else if (c == '"') {
        buf.setCharAt(i, '&');
        buf.insert(i + 1, "#034;");
        i += 5;
      } else if (c == '\'') {
        buf.setCharAt(i, '&');
        buf.insert(i + 1, "#039;");
        i += 5;
      }
    }
    return buf.toString();
  }
}
